
#ifndef FILEBROWSING_EXPORT_H
#define FILEBROWSING_EXPORT_H

#ifdef FILEBROWSING_STATIC_DEFINE
#  define FILEBROWSING_EXPORT
#  define FILEBROWSING_NO_EXPORT
#else
#  ifndef FILEBROWSING_EXPORT
#    ifdef MauiKitFileBrowsing4_EXPORTS
        /* We are building this library */
#      define FILEBROWSING_EXPORT __attribute__((visibility("default")))
#    else
        /* We are using this library */
#      define FILEBROWSING_EXPORT __attribute__((visibility("default")))
#    endif
#  endif

#  ifndef FILEBROWSING_NO_EXPORT
#    define FILEBROWSING_NO_EXPORT __attribute__((visibility("hidden")))
#  endif
#endif

#ifndef FILEBROWSING_DEPRECATED
#  define FILEBROWSING_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef FILEBROWSING_DEPRECATED_EXPORT
#  define FILEBROWSING_DEPRECATED_EXPORT FILEBROWSING_EXPORT FILEBROWSING_DEPRECATED
#endif

#ifndef FILEBROWSING_DEPRECATED_NO_EXPORT
#  define FILEBROWSING_DEPRECATED_NO_EXPORT FILEBROWSING_NO_EXPORT FILEBROWSING_DEPRECATED
#endif

/* NOLINTNEXTLINE(readability-avoid-unconditional-preprocessor-if) */
#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef FILEBROWSING_NO_DEPRECATED
#    define FILEBROWSING_NO_DEPRECATED
#  endif
#endif

#endif /* FILEBROWSING_EXPORT_H */
